import numpy as np
import hyperchamber as hc
import inspect
import collections

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

cg_state = collections.namedtuple("CGState", ["i", "x", "r", "p", "rdotr", "alpha"])
def update_vars(state1, state2):
  ops = []
  for name in state1._fields:
    state1_vs = getattr(state1, name)
    if isinstance(state1_vs, list):
      ops += [tf.assign(_v1, _v2) for _v1, _v2 in zip(state1_vs, getattr(state2, name))]
    else:
      ops += [tf.assign(state1_vs, getattr(state2, name))]
  return tf.group(*ops)

def build_vars(state):
  args = []
  variables = []
  for name in state._fields:
    vs = getattr(state, name)
    if name == 'alpha':
        sv = tf.Variable(vs, trainable=False, name=name+"_sv_dontsave")
        variables += [sv]
    elif isinstance(vs, list):
        sv = [tf.Variable(tf.zeros_like(v), trainable=False, name=name+"_sv_dontsave") for v in vs]
        variables += sv
    else:
        sv = tf.Variable(tf.zeros_like(vs), trainable=False, name=name+"_sv_dontsave")
        variables += [sv]
    args.append(sv)
  return cg_state(*args), variables

def tf_conjugate_gradient(operator,
                       rhs,
                       tol=1e-4,
                       max_iter=20,
                       alpha=1.0,
                       config={},
                       loss=None,
                       params=None,
                       name="conjugate_gradient"):
    r"""
        modified from tensorflow/contrib/solvers/python/ops/linear_equations.py
    """
    def dot(x, y):
      x = [tf.reshape(_x, [-1]) for _x in x]
      y = [tf.reshape(_y, [-1]) for _y in y]
      x = tf.concat(x, axis=0)
      y = tf.concat(y, axis=0)
      return tf.reduce_sum(tf.multiply(x,y))

    def cg_step(state):  # pylint: disable=missing-docstring
      #z = [p * h2 for p, h2 in zip(state.p, operator.apply(state.p))]
      grad = operator.apply(state.p)
      if config.z == "p+grad":
        z = [_grad + _p for _grad, _p in zip(grad, state.p)]
      else:
        z = grad

      if config.alpha == "hager_zhang":
          ValueAndGradient = collections.namedtuple('ValueAndGradient', ['x', 'f', 'df'])
          def value_and_gradients_function(x):
              return ValueAndGradient(x=params, f=loss, df=tf.gradients(loss, params))
          ls_result = tfp.optimizer.linesearch.hager_zhang(value_and_gradients_function, initial_step_size=1.0)
          alpha = ls_result.left
      else:
          #alpha = dot(z,z) / (dot(state.p, state.p)+1e-32)
          alpha = state.alpha#tf.Print(state.alpha, [state.alpha], "Alpha:")

      #x = [_alpha * _p + _x for _alpha, _p, _x in zip(alpha, state.p, state.x)]
      print("X", state.x)
      print("P", state.p)
      print("alpha", alpha)
      x = [alpha * _p + _x for  _p, _x in zip(state.p, state.x)]
      r_1 = state.r
      r = grad#z#[alpha * _z + _r for _z, _r in zip(z, state.r)]

      new_rdotr = dot(r, r)
      beta_fr = new_rdotr / (rdotr+1e-32)
      y = [_r - _p for _r, _p in zip(r, state.p)]
      beta_pk = tf.nn.relu(dot(r,  y) / (dot(state.p, state.p)+1e-32))
      beta_bycd = dot(r, r) / tf.maximum( dot( state.p, y ), dot( [-_p for _p in state.p], r_1 ) )
      p = [_r + beta_bycd * _p for _r, _p in zip(r,state.p)]
      i = state.i + 1

      return cg_state(i, x, r, p, new_rdotr, alpha)

    with tf.name_scope(name):
      x = [tf.zeros_like(h) for h in rhs]
      rdotr = dot(rhs, rhs)
      #p = [-_p for _p in rhs]
      p = rhs
      state = cg_state(i=0, x=x, r=p, p=p, rdotr=rdotr, alpha=alpha)
      state, variables = build_vars(state)
      def update_op(state):
        return update_vars(state, cg_step(state))
      def reset_op(state, rhs):
        return update_vars(state, cg_step(cg_state(i=0, x=x, r=[state.alpha * _g for _g in rhs], p=[state.alpha * _g for _g in rhs], rdotr=rdotr, alpha=state.alpha)))
      return [reset_op(state, rhs), update_op(state), variables, state]

class CGOperator:
    def __init__(self, hvp, x_loss, y_loss, x_params, y_params, lr):
        self.hvp = hvp
        self.x_loss = x_loss
        self.y_loss = y_loss
        self.x_params = x_params
        self.y_params = y_params
        self.lr = lr

    def apply(self, p):
        lr_x = self.lr#tf.sqrt(self.lr)
        lr_y = self.lr
        h_1_v = self.hvp(self.x_loss, self.y_params, self.x_params, [lr_x * _p for _p in p])
        for _x, _h in zip(self.x_params, h_1_v):
            if _h is None:
                print("X none", _x)
        h_2_v = self.hvp(self.y_loss, self.x_params, self.y_params, [lr_y * _h for _h in h_1_v])
        return h_2_v
        #return [lr_x * _g for _g in h_2_v]

class CompetitiveTrainer(BaseTrainer):
    def hessian_vector_product(self, ys, xs, xs2, vs, grads=None):
        if len(vs) != len(xs):
            raise ValueError("xs and v must have the same length.")

        if grads is None:
            grads = tf.gradients(ys, xs)

        assert len(grads) == len(xs)
        elemwise_products = [
                math_ops.multiply(grad_elem, tf.stop_gradient(v_elem))
                for grad_elem, v_elem in zip(grads, vs)
                if grad_elem is not None
                ]

        return tf.gradients(elemwise_products, xs2)

    def _create(self):
        gan = self.gan
        config = self.config
        lr = self.config.learn_rate

        loss = self.gan.loss
        d_loss, g_loss = loss.sample

        config.optimizer["loss"] = loss.sample

        self.optimizer = self.gan.create_optimizer(config.optimizer)

        d_grads = tf.gradients(d_loss, gan.d_vars())
        g_grads = tf.gradients(g_loss, gan.g_vars())

        self.g_loss = g_loss
        self.d_loss = d_loss
        self.gan.trainer = self

        d_params = gan.d_vars()
        g_params = gan.g_vars()
        clarified_d_grads = [tf.Variable(tf.zeros_like(v), trainable=False, name=v.name.split(":")[0]+"_sv_dontsave") for v in d_grads]
        clarified_g_grads = [tf.Variable(tf.zeros_like(v), trainable=False, name=v.name.split(":")[0]+"_sv_dontsave") for v in g_grads]

        clarified_grads = clarified_d_grads + clarified_g_grads
        operator_g = CGOperator(hvp=self.hessian_vector_product, x_loss=d_loss, y_loss=g_loss, x_params=d_params, y_params=g_params, lr=lr)
        reset_g_op, cg_g_op, var_g, state_g = tf_conjugate_gradient( operator_g, clarified_g_grads, max_iter=(self.config.nsteps or 10), config=(self.config.conjugate or hc.Config({})), alpha=(self.config.initial_alpha or 1.0), loss=g_loss, params=g_params)
        operator_d = CGOperator(hvp=self.hessian_vector_product, x_loss=g_loss, y_loss=d_loss, x_params=g_params, y_params=d_params, lr=lr)
        reset_d_op, cg_d_op, var_d, state_d = tf_conjugate_gradient( operator_d, clarified_d_grads, max_iter=(self.config.nsteps or 10), config=(self.config.conjugate or hc.Config({})), alpha=(self.config.initial_alpha or 1.0), loss=d_loss, params=d_params)
        self._variables = var_g + var_d + clarified_g_grads + clarified_d_grads

        assign_d = [tf.assign(c, x) for c, x in zip(clarified_d_grads, d_grads)]
        assign_g = [tf.assign(c, y) for c, y in zip(clarified_g_grads, g_grads)]
        self.reset_clarified_gradients = tf.group(*(assign_d+assign_g))

        self.reset_conjugate_tracker = tf.group(reset_g_op, reset_d_op)
        self.conjugate_gradient_descend_t_1 = tf.group(cg_g_op, cg_d_op)

        assign_d2 = [tf.assign(c, x) for c, x in zip(clarified_d_grads, state_d.x)]
        assign_g2 = [tf.assign(c, y) for c, y in zip(clarified_g_grads, state_g.x)]

        self.conjugate_gradient_descend_t_2 = tf.group(*(assign_d2+assign_g2))
        self.gan.add_metric('cg_g', sum([ tf.reduce_sum(tf.abs(_p)) for _p in clarified_g_grads]))

        if self.config.sga_lambda:
            dyg = tf.gradients(g_loss, g_params)
            dxf = tf.gradients(d_loss, d_params)
            hyp_d = self.hessian_vector_product(d_loss, g_params, d_params, [self.config.sga_lambda * _g for _g in dyg])
            hyp_g = self.hessian_vector_product(g_loss, d_params, g_params, [self.config.sga_lambda * _g for _g in dxf])
            sga_g_op = [tf.assign_sub(_g, _h) for _g, _h in zip(clarified_g_grads, ([state_g.alpha * _g for _g in hyp_g]))]
            sga_d_op = [tf.assign_sub(_g, _h) for _g, _h in zip(clarified_d_grads, ([state_d.alpha * _g for _g in hyp_d]))]
            self.sga_step_t = tf.group(*(sga_d_op + sga_g_op))
            self.gan.add_metric('hyp_g', sum([ tf.reduce_mean(_p) for _p in hyp_g]))
            self.gan.add_metric('hyp_d', sum([ tf.reduce_mean(_p) for _p in hyp_d]))

        #self.clarification_metric_g = sum(state_g.rdotr)
        #self.clarification_metric_d = sum(state_d.rdotr)
        def _metric(r):
            #return tf.reduce_max(tf.convert_to_tensor([tf.reduce_max(tf.norm(_r)) for _r in r]))
            return [tf.reduce_max(tf.norm(_r)) for _r in r][0]
        self.clarification_metric_g = state_g.rdotr
        self.clarification_metric_d = state_d.rdotr

        all_vars = d_params + g_params
        new_grads_and_vars = list(zip(clarified_grads, all_vars)).copy()
    
        self.last_mx = 1e12
        self.last_my = 1e12

        self.alpha = state_g.alpha
        self.alpha_value= tf.constant(1.0)
        set_alpha_d = tf.assign(state_d.alpha, self.alpha_value)
        set_alpha_g = tf.assign(state_g.alpha, self.alpha_value)
        self.set_alpha = tf.group(set_alpha_d, set_alpha_g)
        self.optimize_t = self.optimizer.apply_gradients(new_grads_and_vars)

    def required(self):
        return "".split()

    def variables(self):
        return super().variables() + self._variables

    def _step(self, feed_dict):
        gan = self.gan
        sess = gan.session
        config = self.config
        loss = gan.loss
        metrics = gan.metrics()

        d_loss, g_loss = loss.sample

        self.before_step(self.current_step, feed_dict)
        sess.run(self.reset_clarified_gradients, feed_dict)
        sess.run(self.reset_conjugate_tracker, feed_dict)
        i=0
        sess.run(self.set_alpha, {self.alpha_value: (self.config.initial_alpha or 1.0)})
        alpha = sess.run(self.alpha)

        if self.config.sga_lambda:
            sess.run(self.sga_step_t, feed_dict)

        while True:
            i+=1
            mx, my, _ = sess.run([self.clarification_metric_d, self.clarification_metric_g, self.conjugate_gradient_descend_t_1], feed_dict)
            if i == 1:
                initial_clarification_metric_g = my
                initial_clarification_metric_d = mx
                dy = my
                dx = mx
                my_t1 = my
                mx_t1 = mx
            else:
                dy = my_t1 - my
                dx = mx_t1 - mx
                my_t1 = my
                mx_t1 = mx

            if self.config.log_level == "info":
                alpha = sess.run(self.alpha)
                print("-MD %e MG %e alpha %e" % (mx, my, alpha))

            if mx > initial_clarification_metric_d or my > initial_clarification_metric_g or np.isnan(mx) or np.isnan(my) or np.isinf(mx) or np.isinf(my):
                sess.run(self.set_alpha, {self.alpha_value: alpha / 2.0})
                sess.run(self.reset_clarified_gradients, feed_dict)
                sess.run(self.reset_conjugate_tracker, feed_dict)
                if self.config.sga_lambda:
                    sess.run(self.sga_step_t, feed_dict)
                new_alpha = sess.run(self.alpha)
                if new_alpha < 1e-10:
                    print("Alpha too low, exploding")
                    break
                print(i, "Explosion detected, reduced alpha from ", alpha, "to", new_alpha)
                alpha = new_alpha
                i=0
                continue

            threshold_x =  mx / (initial_clarification_metric_d+1e-12)
            threshold_y = my / (initial_clarification_metric_g+1e-12)
            threshold = threshold_x + threshold_y

            if self.config.max_steps and i > self.config.max_steps:
               if self.config.verbose:
                   print("Max steps ", self.config.max_steps, "threshold", threshold, "max", self.config.threshold, "mx", mx, "my", my)
               break
            #print( "p", i, dx / mx, dy / my, self.config.threshold, dy / my > self.config.threshold)
            if i % 100 == 0 and i != 0:
                print("Threshold at", i, threshold_x, threshold_y, "last", self.last_mx, self.last_my, "m", mx, my, "d", dx, dy)
            #self.last_mx = (self.last_mx + 1e-16) * 1.05
            #self.last_my = (self.last_my + 1e-16) * 1.05
            #if ((dx / (initial_clarification_metric_d+1e-12)) < self.config.threshold and (dy / (initial_clarification_metric_g+1e-12)) < self.config.threshold and my <= self.last_my and mx <= self.last_mx):
            if mx < self.config.threshold and my < self.config.threshold:
                #self.last_mx = mx
                #self.last_my = my
                sess.run(self.conjugate_gradient_descend_t_2)
                if self.config.verbose:
                    print("Found in ", i, "threshold", threshold, "mx", mx, "my", my, "alpha", alpha)
                break
        metric_values = sess.run([self.optimize_t] + self.output_variables(metrics), feed_dict)[1:]
        self.after_step(self.current_step, feed_dict)

        if self.current_step % 10 == 0:
            print(str(self.output_string(metrics) % tuple([self.current_step] + metric_values)))




import hypergan as hg
import hyperchamber as hc
import numpy as np
import random
import copy

from hypergan.losses.wasserstein_loss import WassersteinLoss
from hypergan.losses.least_squares_loss import LeastSquaresLoss

class RandomSearch:
    def __init__(self, overrides):
        self.options = {
            'discriminator': self.discriminator(),
            'generator': self.generator(),
            'trainer': self.trainer(),
            'loss':self.loss(),
            'encoder':self.encoder()
         }

        self.options = {**self.options, **overrides}

    def range(self, start=0., end=1.):
        return list(np.linspace(start, end, num=1000))

    def trainer(self):
        tftrainers = [
                #tf.train.AdadeltaOptimizer,
                #tf.train.AdagradOptimizer,
                #tf.train.GradientDescentOptimizer,
                tf.train.AdamOptimizer,
                #tf.train.MomentumOptimizer,
                tf.train.RMSPropOptimizer
        ]

        selector = hc.Selector({
            'learn_rate': [1e-2, 1e-3, 1e-4, 5e-3, 5e-4],
            'beta1': self.range(0.8, 0.9999),
            'beta2': self.range(0.9, 0.9999),
            'epsilon': self.range(1e-8, 0.1),
            'momentum': [0, 0.01, 0.1],
            'decay': self.range(0.8, 0.9999),
            'rho': self.range(),
            'initial_accumulator_value': self.range(),
            'clipped_gradients': False,
            'trainer':tftrainers,
            'class': [
                #hg.trainers.proportional_control_trainer.create,
                #hg.trainers.alternating_trainer.AlternatingTrainer
                hg.trainers.consensus_trainer.ConsensusTrainer
            ]
        })
        
        config = selector.random_config()
        return config
     
    def fc_discriminator(self):
        opts = {
          "activation": ["selu", "lrelu", "relu"],
          "layer_regularizer": [None, "layer_norm"],
          "linear_type": [None, "cosine"],
          "features": [1, 10, 100, 200, 512],
          "class": "class:hypergan.discriminators.fully_connected_discriminator.FullyConnectedDiscriminator"
        }
        return hc.Selector(opts).random_config()

    def var_loss(self):
        loss_opts = {
            'class': [
                    VralLoss
            ],
            "target_mean": [-1,-0.5,0,0.5,1],
            "fake_mean": [-1,-0.5,0,0.5,1],
            'reduce': ['reduce_mean','reduce_sum','reduce_logsumexp'],
            'type': ['log_rr', 'log_rf', 'log_fr', 'log_ff', 'log_all'],
            'value_function': ['square', 'log', 'original'],
            'g_loss': ['l2','fr_l2','rr_l2'],
            
            "r_discriminator": self.fc_discriminator()

        }
        loss_opts["f_discriminator"] = loss_opts["r_discriminator"]

        return  hc.Selector(loss_opts).random_config()

    def loss(self):
        a=self.loss_instance()
        b=copy.deepcopy(dict(a))
        b["swapped"]=True
        
        loss={
            "class": "class:hypergan.losses.multi_loss.MultiLoss",
            "combine": "concat",
            "partition": True
        }

        loss["losses"]=[a,b]
        return loss

    def loss_instance(self):
        loss_opts = {
            'class': [
                    FDivergenceLoss, StandardLoss, LeastSquaresLoss, WassersteinLoss
            ],
            "type": ["kl","js","gan","reverse_kl","pearson","squared_hellinger", "total_variation"],
            "labels": [[-1,1,1]],
            'reduce': ['reduce_mean']#,'reduce_sum']#,'reduce_logsumexp']
        }

        choice = hc.Selector(loss_opts).random_config()

        if random.choice([True, False]):
            choice["regularizer"] = choice["type"]
        if random.choice([True, False]):
            choice["g_loss_type"] = choice["type"]

        return choice


    def encoder(self):
        projections = []
        projections.append([hg.encoders.uniform_distribution.identity])
        encoder_opts = {
                'z': 1,
                'min': -1,
                'max':1,
                "projections": projections,
                'class': hg.encoders.uniform_distribution.UniformDistribution
        }

        return hc.Selector(encoder_opts).random_config()

    def generator(self):
        generator_opts = {
            "activation":['lrelu', 'tanh', 'selu', 'prelu', 'crelu', 'nsoftplus'],
            "final_depth":[32],
            "depth_increase":[32],
            "initializer": ['xavier'],
            "random_stddev": list(np.linspace(0.0, 0.1, num=10000)),
            "final_activation":['lrelu', 'tanh', None],
            "block_repeat_count":[1,2,3],
            "block":[
                hg.generators.common.standard_block, 
                hg.generators.common.inception_block, 
                hg.generators.common.dense_block, 
                hg.generators.common.repeating_block
                ],
            "orthogonal_initializer_gain": list(np.linspace(0.1, 2, num=100)),
            "class":[
                hg.generators.resize_conv_generator.ResizeConvGenerator
            ]
        }

        return hc.Selector(generator_opts).random_config()

    def discriminator(self):
        discriminator_opts = {
            "activation":['relu', 'lrelu', 'tanh', 'selu', 'prelu', 'crelu'],
            "final_activation":[None],
            "block_repeat_count":[1,2,3],
            "block":[hg.discriminators.common.repeating_block,
                   hg.discriminators.common.standard_block,
                   hg.discriminators.common.strided_block
                   ],
            "depth_increase":[32],
            "extra_layers": [0, 1, 2, 3],
            "extra_layers_reduction":[1,2,4],
            "fc_layer_size":[300, 400, 500],
            "fc_layers":[0,1],
            "first_conv_size":[32],
            "layers": [3,4,5,6],
            "initial_depth": [32],
            "initializer": ['xavier'],
            "layer_regularizer": [None,  'layer_norm'],
            "noise":[False, 1e-2],
            "progressive_enhancement":[False, True],
            "orthogonal_gain": list(np.linspace(0.1, 2, num=10000)),
            "random_stddev": list(np.linspace(0.0, 0.1, num=10000)),
            "distance":['l1_distance', 'l2_distance'],
            "class":[
                hg.discriminators.pyramid_discriminator.PyramidDiscriminator
               # hg.discriminators.autoencoder_discriminator.AutoencoderDiscriminator
            ]
        }

        return hc.Selector(discriminator_opts).random_config()

    def random_config(self):
        return hc.Selector(self.options).random_config()



#From https://gist.github.com/EndingCredits/b5f35e84df10d46cfa716178d9c862a3
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import hyperchamber as hc
import numpy as np
import inspect
from operator import itemgetter
from hypergan.train_hooks.base_train_hook import BaseTrainHook

class RollingMemoryTrainHook(BaseTrainHook):
  "Keeps a rolling memory of the best scoring discriminator samples."
  def __init__(self, gan=None, config=None, trainer=None, name="RollingMemoryTrainHook"):
    super().__init__(config=config, gan=gan, trainer=trainer, name=name)
    config = hc.Config(config)
    s = self.gan.ops.shape(self.gan.generator.sample)
    self.shape = s#[self.gan.batch_size() * (self.config.memory_size or 1), s[1], s[2], s[3]]
    with tf.variable_scope((self.config.name or self.name), reuse=self.gan.reuse) as scope:
        self.mx=tf.get_variable(self.gan.ops.generate_name()+"_dontsave", s, dtype=tf.float32,
                  initializer=tf.compat.v1.constant_initializer(-100), aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA, trainable=False)
        self.mg=tf.get_variable(self.gan.ops.generate_name()+"_dontsave", s, dtype=tf.float32,
                  initializer=tf.compat.v1.constant_initializer(100), aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA, trainable=False)
    self.m_discriminator = gan.create_component(gan.config.discriminator, name="discriminator", input=tf.concat([self.mx, self.mg],axis=0), features=[gan.features], reuse=True)
    self.m_loss = gan.create_component(gan.config.loss, discriminator=self.m_discriminator)
    swx = self.m_loss.d_real
    swg = self.m_loss.d_fake
    if self.config.reverse_mx:
        swx = -swx
    if self.config.reverse_mg:
        swg = -swg
    swx = tf.reshape(swx, [-1])
    swg = tf.reshape(swg, [-1])
    _, swx = tf.nn.top_k(swx, k=(self.config.top_k or 1), sorted=True, name=None)
    _, swg = tf.nn.top_k(swg, k=(self.config.top_k or 1), sorted=True, name=None)
    swx = tf.one_hot(swx, self.gan.batch_size(), dtype=tf.float32)
    swg = tf.one_hot(swg, self.gan.batch_size(), dtype=tf.float32)
    swx = tf.reduce_sum(swx, reduction_indices=0)
    swg = tf.reduce_sum(swg, reduction_indices=0)
    swx = tf.reshape(swx, [self.gan.batch_size(), 1, 1, 1])
    swg = tf.reshape(swg, [self.gan.batch_size(), 1, 1, 1])
    self.swx = swx
    self.swg = swg
    self.assign_mx = tf.assign(self.mx, self.gan.inputs.x * swx + (1.0 - swx) * self.mx)
    self.assign_mg = tf.assign(self.mg, self.gan.generator.sample * swg + (1.0 - swg) * self.mg)
    self.assign_ops = tf.group(*[self.assign_mx, self.assign_mg])
    self.train_hook_index = len(trainer.train_hooks)
    self.loss = [tf.zeros(1), tf.zeros(1)]
    for _type in self.config.types or ['mx/mg']:
        if _type == 'mg/g':
            self.mg_discriminator = gan.create_component(gan.config.discriminator, name="discriminator", input=tf.concat([self.mg, self.gan.generator.sample],axis=0), features=[gan.features], reuse=True)
            self.mg_loss = gan.create_component(gan.config.loss, discriminator=self.mg_discriminator)
            self.gan.losses += [self.mg_loss]
            self.loss[0] += (self.config.lam or 1.0) * self.mg_loss.sample[0]
            self.loss[1] += (self.config.lam or 1.0) * self.mg_loss.sample[1]
            self.gan.add_metric('roll_loss_mg/g', self.loss[0])
        elif _type == 'mx/mg': 
            self.loss[0] += (self.config.lam or 1.0) * self.m_loss.sample[0]
            self.loss[1] += (self.config.lam or 1.0) * self.m_loss.sample[1]
            self.gan.add_metric('roll_loss_mx/mg', self.loss[0])
        elif _type == 'mx/g': 
            self.mg_discriminator = gan.create_component(gan.config.discriminator, name="discriminator", input=tf.concat([self.mx, self.gan.generator.sample],axis=0), features=[gan.features], reuse=True)
            self.mg_loss = gan.create_component(gan.config.loss, discriminator=self.mg_discriminator)
            self.loss[0] += (self.config.lam or 1.0) * self.mg_loss.sample[0]
            self.loss[1] += (self.config.lam or 1.0) * self.mg_loss.sample[1]
            self.gan.add_metric('roll_loss_mx/g', self.loss[0])
        elif _type == 'x/mg': 
            self.mg_discriminator = gan.create_component(gan.config.discriminator, name="discriminator", input=tf.concat([self.gan.inputs.x, self.mg],axis=0), features=[gan.features], reuse=True)
            self.mg_loss = gan.create_component(gan.config.loss, discriminator=self.mg_discriminator)
            self.loss[0] += (self.config.lam or 1.0) * self.mg_loss.sample[0]
            self.loss[1] += (self.config.lam or 1.0) * self.mg_loss.sample[1]
            self.gan.add_metric('roll_loss_x/mg', self.loss[0])
        elif _type == 'x/mx': 
            self.mg_discriminator = gan.create_component(gan.config.discriminator, name="discriminator", input=tf.concat([self.mg, self.gan.generator.sample],axis=0), features=[gan.features], reuse=True)
            self.mg_loss = gan.create_component(gan.config.loss, discriminator=self.mg_discriminator)
            self.loss[0] += (self.config.lam or 1.0) * self.mg_loss.sample[0]
            self.loss[1] += (self.config.lam or 1.0) * self.mg_loss.sample[1]
            self.gan.add_metric('roll_loss_x/mx', self.loss[0])


  def distributed_step(self, input_iterator_next):
    def assign_mx(mx, inp, swx):
        inp2 = self.gan.replica.inputs.x
        op=self.gan.replica.trainer.train_hooks[self.train_hook_index].assign_mx
        with tf.control_dependencies([op]):
            return tf.no_op()
    def assign_mg(mg, gen, swg):
        op = self.gan.replica.trainer.train_hooks[self.train_hook_index].assign_mg
        with tf.control_dependencies([op]):
            return tf.no_op()
    mxop = self.gan.distribution_strategy.extended.call_for_each_replica(assign_mx, args=(self.mx, input_iterator_next, self.swx,))
    mgop = self.gan.distribution_strategy.extended.call_for_each_replica(assign_mg, args=(self.mg, self.gan.generator.sample, self.swg,))

    return [mxop, mgop]

  def distributed_debug(self):
    mxop = self.gan.distribution_strategy.extended.read_var(self.mx)
    mgop = self.gan.distribution_strategy.extended.read_var(self.mg)

    return [mxop, mgop]

  def distributed_initial_step(self, input_iterator_next):
    def assign_mx(mx, inp):
        return mx.assign(inp)

    def assign_mg(mg, gen):
        return mg.assign(gen)

    mxop = self.gan.distribution_strategy.extended.call_for_each_replica(assign_mx, args=(self.mx,input_iterator_next,))
    mgop = self.gan.distribution_strategy.extended.call_for_each_replica(assign_mg, args=(self.mg,self.gan.generator.sample,))
    return [mxop, mgop]


  def before_step(self, step, feed_dict):
      if step == 0:
          self.gan.session.run(tf.assign(self.mx, self.gan.inputs.x))
          self.gan.session.run(tf.assign(self.mg, self.gan.generator.sample))

  def after_step(self, step, feed_dict):
      self.gan.session.run(self.assign_ops)

  def variables(self):
      return [self.mx, self.mg]

  def losses(self):
    return self.loss



from .base_gan import BaseGAN
from hyperchamber import Config
from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.gan_component import ValidationException, GANComponent
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *
import copy
import hyperchamber as hc
import hypergan as hg
import importlib
import json
import numpy as np
import os
import sys
import time
import torch
import uuid


class AlignedInterpolateGAN(BaseGAN):
    """ 
    """
    def __init__(self, *args, **kwargs):
        BaseGAN.__init__(self, *args, **kwargs)

    def create(self):
        self.latent = self.create_component("latent")
        self.x = self.inputs.next()[0]
        if self.config.use_latent:
            self.generator = self.create_component("generator", input_shape=[self.latent.next().shape[1]])
        else:
            self.generator = self.create_component("generator", input=self.x)
        if self.config.shared_discriminator:
            self.discriminator = self.create_component("discriminator", context_shapes={"class": [1]})
        else:
            self.discriminator = self.create_component("discriminator")
            self.discriminator2 = self.create_component("discriminator")
        self.loss = self.create_component("loss")
        self.trainer = self.create_component("trainer")
        self.sigmoid = torch.nn.Sigmoid()
        self.gammas = [torch.Tensor([self.config.interpolate]).float()[0].cuda(), torch.Tensor([1.-self.config.interpolate]).float()[0].cuda()]

    def forward_discriminator(self, inputs):
        if self.config.shared_discriminator:
            d0_class = torch.zeros([inputs[0].shape[0], 1], device="cuda:0")
            d1_class = torch.ones([inputs[1].shape[0], 1], device="cuda:0")
            d0 = self.discriminator(inputs[0], context={"class": d0_class})
            d1 = self.discriminator(inputs[1], context={"class": d1_class})
            return d0 * self.gammas[0] * torch.sigmoid(d1) + d1 * self.gammas[1] * torch.sigmoid(d0)

        if self.config.union:
            d0 = self.discriminator(inputs[0])
            d1 = self.discriminator2(inputs[1])
            d3 = self.discriminator(inputs[1])
            d4 = self.discriminator2(inputs[0])
            return d0 * self.gammas[0] * torch.sigmoid(d1) + d1 * self.gammas[1] * torch.sigmoid(d0) + \
                    d3 * self.gammas[0] * torch.sigmoid(d4) + d4 * self.gammas[1] * torch.sigmoid(d3)

        else:
            d0 = self.discriminator(inputs[0])
            d1 = self.discriminator2(inputs[1])
            return self.sigmoid(d1)*d0*self.gammas[0]  + self.sigmoid(d0)*d1*self.gammas[1]

    def forward_pass(self):
        self.x = self.inputs.next()
        self.y = self.inputs.next(1)
        if self.config.use_latent:
            g = self.generator(self.latent.next())
        else:
            g = self.generator(self.x)
        self.g = g
        d_real = self.forward_discriminator([self.x, self.y])
        d_fake = self.forward_discriminator([self.g, self.g])
        self.d_fake = d_fake
        self.d_real = d_real
        return d_real, d_fake

    def forward_loss(self):
        d_real, d_fake= self.forward_pass()
        d_loss, g_loss = self.loss.forward(d_real, d_fake)
        return d_loss, g_loss

    def discriminator_components(self):
        if self.config.shared_discriminator:
            return [self.discriminator]
        else:
            return [self.discriminator, self.discriminator2]

    def generator_components(self):
        return [self.generator]

    def discriminator_fake_inputs(self, discriminator_index=0):
        return [self.g, self.g]

    def discriminator_real_inputs(self, discriminator_index=0):
        if hasattr(self, 'y'):
            return [self.x, self.y]
        else:
            return [self.inputs.next(), self.inputs.next(1)]



import numpy as np
import hyperchamber as hc
import inspect

from hypergan.trainers.base_trainer import BaseTrainer

TINY = 1e-12

class IncrementalTrainer():
    def __init__(self, trainers, schedule):
        self.trainers = trainers
        self.schedule = schedule
        self.i = 0

    def step(self, feed_dict):
        self.i += 1

        index = 0
        threshold = 0
        for val in self.schedule:
            threshold += val
            if self.i > threshold:
                index+=1
                if len(self.trainers) <= index:
                    index =len(self.trainers) - 1

        return self.trainers[index].step(feed_dict)






import importlib
import json
import numpy as np
import os
import sys
import time
import uuid
import copy

from hypergan.discriminators import *
from hypergan.distributions import *
from hypergan.generators import *
from hypergan.inputs import *
from hypergan.samplers import *
from hypergan.trainers import *

import hyperchamber as hc
from hyperchamber import Config
from hypergan.ops import TensorflowOps
import hypergan as hg

from hypergan.gan_component import ValidationException, GANComponent
from ..base_gan import BaseGAN

from hypergan.distributions.uniform_distribution import UniformDistribution
from hypergan.trainers.experimental.consensus_trainer import ConsensusTrainer

class ConditionalGAN(BaseGAN):
    """ 
    """
    def __init__(self, *args, **kwargs):
        BaseGAN.__init__(self, *args, **kwargs)

    def required(self):
        """
        `input_encoder` is a discriminator.  It encodes X into Z
        `discriminator` is a standard discriminator.  It measures X, reconstruction of X, and G.
        `generator` produces two samples, input_encoder output and a known random distribution.
        """
        return "generator discriminator ".split()

    def create(self):
        config = self.config
        ops = self.ops

        with tf.device(self.device):
            #x_input = tf.identity(self.inputs.x, name='input')
            xa_input = tf.identity(self.inputs.xa, name='xa_i')
            xb_input = tf.identity(self.inputs.xb, name='xb_i')

            def random_like(x):
                return UniformDistribution(self, config.z_distribution, output_shape=self.ops.shape(x)).sample
            #y=a
            #x=b
            zgx = self.create_component(config.encoder, input=xa_input, name='xa_to_x')
            zgy = self.create_component(config.encoder, input=xb_input, name='xb_to_y')
            zx = zgx.sample
            zy = zgy.sample

            z_noise = random_like(zx)
            n_noise = random_like(zx)
            if config.style:
                stylex = self.create_component(config.style_discriminator, input=xb_input, name='xb_style')
                styley = self.create_component(config.style_discriminator, input=xa_input, name='xa_style')
                zy = tf.concat(values=[zy, z_noise], axis=3)
                zx = tf.concat(values=[zx, n_noise], axis=3)
                gy = self.create_component(config.generator, features=[styley.sample], input=zy, name='gy_generator')
                y = hc.Config({"sample": xa_input})
                zx = self.create_component(config.encoder, input=y.sample, name='xa_to_x', reuse=True).sample
                zx = tf.concat(values=[zx, z_noise], axis=3)
                gx = self.create_component(config.generator, features=[stylex.sample], input=zx, name='gx_generator')
            else:
                gy = self.create_component(config.generator, features=[z_noise], input=zy, name='gy_generator')
                y = hc.Config({"sample": xa_input})
                zx = self.create_component(config.encoder, input=y.sample, name='xa_to_x', reuse=True).sample
                gx = self.create_component(config.generator, features=[z_noise], input=zx, name='gx_generator')
                stylex=hc.Config({"sample":random_like(y.sample)})

            self.y = y
            self.gy = gy
            self.gx = gx

            ga = gy
            gb = gx

            self.uniform_sample = gb.sample

            xba = ga.sample
            xab = gb.sample
            xa_hat = ga.reuse(zx)
            xb_hat = gb.reuse(zy)
            xa = xa_input
            xb = xb_input

            self.styleb = stylex
            self.random_style = random_like(stylex.sample)



            t0 = xb
            t1 = gx.sample
            f0 = gy.sample
            f1 = y.sample
            stack = [t0, t1]
            stacked = ops.concat(stack, axis=0)
            features = ops.concat([f0, f1], axis=0)
            self.inputs.x = xa
            ugb = gb.reuse(random_like(zy))
            zub = zy
            sourcezub = zy


            d = self.create_component(config.discriminator, name='d_ab', 
                    input=stacked, features=[features])
            l = self.create_loss(config.loss, d, xa_input, ga.sample, len(stack))
            loss1 = l
            d_loss1 = l.d_loss
            g_loss1 = l.g_loss

            d_vars1 = d.variables()
            g_vars1 = gb.variables()+ga.variables()+zgx.variables()+zgy.variables()

            d_loss = l.d_loss
            g_loss = l.g_loss


            metrics = {
                    'g_loss': l.g_loss,
                    'd_loss': l.d_loss
                }


            trainers = []

            lossa = hc.Config({'sample': [d_loss1, g_loss1], 'metrics': metrics})
            #lossb = hc.Config({'sample': [d_loss2, g_loss2], 'metrics': metrics})
            trainers += [ConsensusTrainer(self, config.trainer, loss = lossa, g_vars = g_vars1, d_vars = d_vars1)]
            #trainers += [ConsensusTrainer(self, config.trainer, loss = lossb, g_vars = g_vars2, d_vars = d_vars2)]
            trainer = MultiTrainerTrainer(trainers)
            self.initialize_variables()

        self.trainer = trainer
        self.generator = gb
        self.encoder = hc.Config({"sample":ugb}) # this is the other gan
        self.uniform_distribution = hc.Config({"sample":zub})#uniform_encoder
        self.uniform_distribution_source = hc.Config({"sample":sourcezub})#uniform_encoder
        self.zb = zy
        self.z_hat = gb.sample
        self.x_input = xa_input
        self.autoencoded_x = xb_hat

        self.cyca = xa_hat
        self.cycb = xb_hat
        self.xba = xba
        self.xab = xab
        self.uga = y.sample
        self.ugb = ugb

        rgb = tf.cast((self.generator.sample+1)*127.5, tf.int32)
        self.generator_int = tf.bitwise.bitwise_or(rgb, 0xFF000000, name='generator_int')


    def create_loss(self, loss_config, discriminator, x, generator, split):
        loss = self.create_component(loss_config, discriminator = discriminator, x=x, generator=generator, split=split)
        return loss

    def create_encoder(self, x_input, name='input_encoder'):
        config = self.config
        input_encoder = dict(config.input_encoder or config.g_encoder or config.generator)
        encoder = self.create_component(input_encoder, name=name, input=x_input)
        return encoder

    def create_z_discriminator(self, z, z_hat):
        config = self.config
        z_discriminator = dict(config.z_discriminator or config.discriminator)
        z_discriminator['layer_filter']=None
        net = tf.concat(axis=0, values=[z, z_hat])
        encoder_discriminator = self.create_component(z_discriminator, name='z_discriminator', input=net)
        return encoder_discriminator

    def create_cycloss(self, x_input, x_hat):
        config = self.config
        ops = self.ops
        distance = config.distance or ops.lookup('l1_distance')
        pe_layers = self.gan.skip_connections.get_array("progressive_enhancement")
        cycloss_lambda = config.cycloss_lambda
        if cycloss_lambda is None:
            cycloss_lambda = 10
        
        if(len(pe_layers) > 0):
            mask = self.progressive_growing_mask(len(pe_layers)//2+1)
            cycloss = tf.reduce_mean(distance(mask*x_input,mask*x_hat))

            cycloss *= mask
        else:
            cycloss = tf.reduce_mean(distance(x_input, x_hat))

        cycloss *= cycloss_lambda
        return cycloss


    def create_z_cycloss(self, z, x_hat, encoder, generator):
        config = self.config
        ops = self.ops
        total = None
        distance = config.distance or ops.lookup('l1_distance')
        if config.z_hat_lambda:
            z_hat_cycloss_lambda = config.z_hat_cycloss_lambda
            recode_z_hat = encoder.reuse(x_hat)
            z_hat_cycloss = tf.reduce_mean(distance(z_hat,recode_z_hat))
            z_hat_cycloss *= z_hat_cycloss_lambda
        if config.z_cycloss_lambda:
            recode_z = encoder.reuse(generator.reuse(z))
            z_cycloss = tf.reduce_mean(distance(z,recode_z))
            z_cycloss_lambda = config.z_cycloss_lambda
            if z_cycloss_lambda is None:
                z_cycloss_lambda = 0
            z_cycloss *= z_cycloss_lambda

        if config.z_hat_lambda and config.z_cycloss_lambda:
            total = z_cycloss + z_hat_cycloss
        elif config.z_cycloss_lambda:
            total = z_cycloss
        elif config.z_hat_lambda:
            total = z_hat_cycloss
        return total



    def input_nodes(self):
        "used in hypergan build"
        if hasattr(self.generator, 'mask_generator'):
            extras = [self.mask_generator.sample]
        else:
            extras = []
        return extras + [
                self.x_input
        ]


    def output_nodes(self):
        "used in hypergan build"

    
        if hasattr(self.generator, 'mask_generator'):
            extras = [
                self.mask_generator.sample, 
                self.generator.g1x,
                self.generator.g2x
            ]
        else:
            extras = []
        return extras + [
                self.encoder.sample,
                self.generator.sample, 
                self.uniform_sample,
                self.generator_int
        ]
